/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <libsyscommon/dbus-system.h>

#include "apps/apps.h"
#include "core/log.h"
#include "battery.h"
#include "config.h"
#include "power-supply.h"
#include "lowbat-handler.h"
#include "battery-ops.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "display/poll.h"
#include "shared/eventsystem.h"
#include "shared/plugin.h"
#include "core/device-notifier.h"


#define METHOD_LOW_NOTI_ON          "BatteryLowNotiOn"
#define METHOD_LOW_NOTI_UPDATE      "BatteryLowNotiUpdate"
#define METHOD_LOW_NOTI_OFF         "BatteryLowNotiOff"
#define METHOD_CRITICAL_NOTI_ON     "BatteryCriticalNotiOn"
#define METHOD_CRITICAL_NOTI_UPDATE "BatteryCriticalNotiUpdate"
#define METHOD_CRITICAL_NOTI_OFF    "BatteryCriticalNotiOff"

#define REMOVE_POPUP        "remove_battery_popups"
#define DISCONNECT_POPUP    "battdisconnect"

enum event_noti_type {
	NOTI_NONE = -1,
	NOTI_UPDATE = 0,
};

static struct display_plugin *disp_plgn;
static struct battery_status *battery;
static guint abnormal_timer;

static int noti_low = NOTI_NONE;
static int noti_crit = NOTI_NONE;

static bool check_remove_condition(const char *prev, const char *cur)
{
	int prev_len;
	int cur_len;
	bool ret = false;

	prev_len = strlen(prev);
	cur_len = strlen(cur);

	if (!prev_len || !cur_len)
		return false;

	if (!strncmp(prev, EVT_VAL_BATTERY_LEVEL_CRITICAL,prev_len)) {
		if (!strncmp(cur, EVT_VAL_BATTERY_LEVEL_LOW, cur_len)
			|| !strncmp(cur, EVT_VAL_BATTERY_LEVEL_HIGH, cur_len)
			|| !strncmp(cur, EVT_VAL_BATTERY_LEVEL_FULL, cur_len))
		   ret = true;
	} else if (!strncmp(prev, EVT_VAL_BATTERY_LEVEL_LOW, prev_len)) {
		if (!strncmp(cur, EVT_VAL_BATTERY_LEVEL_HIGH, cur_len)
			|| !strncmp(cur, EVT_VAL_BATTERY_LEVEL_FULL, cur_len))
			ret = true;
	}

	return ret;
}

static void low_noti_cb(GVariant *var, void *user_data, GError *err)
{

	if (!var) {
		_E("Invalid parameter.");
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &noti_low)) {
		_E("Failed to notify low: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	_D("Inserted battery low noti : %d", noti_low);

out:
	g_variant_unref(var);
}

static void critical_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to notify critical: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	noti_crit = id;
	_D("Inserted battery critical noti : %d", noti_crit);

out:
	g_variant_unref(var);
}

static int launch_lowbat_noti(int capacity, int option)
{
	int ret = -EINVAL;
	char batcap[10] = {0,};
	char notiid[10] = {0,};
	int noti_id;
	const char *pa[2];
	int retry;
	char *noti_type = NULL;
	static int prev_level = BATTERY_NORMAL;

	snprintf(batcap, sizeof(batcap), "%d", capacity);
	pa[0] = batcap;

	if (option == BAT_OPT_WARNING) {
		if (noti_crit > 0) {
			/* remove waring notiid*/
			noti_id = noti_crit;
			for (retry = RETRY_MAX; retry > 0; retry--) {
				ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
						POPUP_PATH_BATTERY,
						POPUP_INTERFACE_BATTERY,
						METHOD_LOW_NOTI_OFF,
						g_variant_new("(i)", noti_id));
				if (ret == 0) {
					_D("Deleted battery critical noti.");
					break;
				} else
					_E("Failed to call dbus method (err: %d)", ret);
			}
		}
		memset(notiid, 0x0, sizeof(notiid));
		snprintf(notiid, sizeof(notiid), "%d", noti_low);
		pa[1] = notiid;
		if (prev_level == battery_info.critical || noti_low == NOTI_UPDATE)
			noti_type = METHOD_LOW_NOTI_UPDATE;
		else {
			noti_type = METHOD_LOW_NOTI_ON;
			if (disp_plgn->pm_change_internal)
				disp_plgn->pm_change_internal(INTERNAL_LOCK_LOWBAT, LCD_NORMAL);
		}
		prev_level = battery_info.warning;
		if (battery->charge_now)
			ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
				POPUP_PATH_BATTERY,
				POPUP_INTERFACE_BATTERY,
				noti_type,
				g_variant_new("(ss)", pa[0], pa[1]));
		else
			ret = dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
				POPUP_PATH_BATTERY,
				POPUP_INTERFACE_BATTERY,
				noti_type,
				g_variant_new("(ss)", pa[0], pa[1]),
				low_noti_cb, -1, NULL);
	} else if (option == BAT_OPT_CRITICAL) {
		if (noti_low > 0) {
			/* remove waring notiid*/
			noti_id = noti_low;
			for (retry = RETRY_MAX; retry > 0; retry--) {
				ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
						POPUP_PATH_BATTERY,
						POPUP_INTERFACE_BATTERY,
						METHOD_CRITICAL_NOTI_OFF,
						g_variant_new("(i)", noti_id));
				if (ret == 0) {
					_D("Deleted battery low noti");
					break;
				} else
					_E("Failed to call dbus method (err: %d)", ret);
			}
		}
		memset(notiid, 0x0, sizeof(notiid));
		snprintf(notiid, sizeof(notiid), "%d", noti_crit);
		pa[1] = notiid;
		if (noti_crit == NOTI_UPDATE)
			noti_type = METHOD_CRITICAL_NOTI_UPDATE;
		else {
			noti_type = METHOD_CRITICAL_NOTI_ON;
			if (disp_plgn->pm_change_internal)
				disp_plgn->pm_change_internal(INTERNAL_LOCK_LOWBAT, LCD_NORMAL);
		}
		prev_level = battery_info.critical;
		if (battery->charge_now)
			ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
				POPUP_PATH_BATTERY,
				POPUP_INTERFACE_BATTERY,
				noti_type,
				g_variant_new("(ss)", pa[0], pa[1]));
		else
			ret = dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
				POPUP_PATH_BATTERY,
				POPUP_INTERFACE_BATTERY,
				noti_type,
				g_variant_new("(ss)", pa[0], pa[1]),
				critical_noti_cb, -1, NULL);

	} else
		prev_level = battery_info.normal;

	_D("Requested type(%s) id(%s)", noti_type, notiid);
	return ret;
}

static void clean_lowbat_noti(const char *prev, const char *str)
{
	int noti_id;
	int retry;
	int ret;
	bool condition;

	if (!prev || !str) {
		_E("There is no data.");
		return;
	}

	if (prev == str && !battery->charge_now)
		return;

	condition = check_remove_condition(prev, str);
	_D("Condition: %d", condition);

	if (condition || battery->charge_now) {
		if (noti_crit > 1) {
			/* remove critical notiid */
			noti_id = noti_crit;
			for (retry = RETRY_MAX; retry > 0; retry--) {
				ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
						POPUP_PATH_BATTERY,
						POPUP_INTERFACE_BATTERY,
						METHOD_CRITICAL_NOTI_OFF,
						g_variant_new("(i)", noti_id));
				if (ret == 0) {
					_D("Deleted battery critical noti.");
					noti_crit = NOTI_NONE;
					break;
				} else
					_E("Failed to call dbus method: %d", ret);
			}
		}

		if (noti_low > 1) {
			/* remove waring notiid*/
			noti_id = noti_low;
			for (retry = RETRY_MAX; retry > 0; retry--) {
				ret = dbus_handle_method_async_var(POPUP_BUS_NAME,
						POPUP_PATH_BATTERY,
						POPUP_INTERFACE_BATTERY,
						METHOD_LOW_NOTI_OFF,
						g_variant_new("(i)", noti_id));
				if (ret == 0) {
					_D("Deleted battery low noti.");
					noti_low = NOTI_NONE;
					break;
				} else
					_E("Failed to call dbus method: %d", ret);
			}
		}
	}
	_D("Noti_crit(%d) noti_low(%d).", noti_crit, noti_low);
}

static int changed_battery_cf(int status)
{
	if (status == PRESENT_ABNORMAL)
		return launch_system_app(APP_ABNORMAL, 2, APP_KEY_TYPE, DISCONNECT_POPUP);
	return launch_system_app(APP_REMOVE, 2, APP_KEY_TYPE, REMOVE_POPUP);
}

static void remove_health_popup(void)
{
	int ret;

	ret = launch_system_app(APP_REMOVE, 2, APP_KEY_TYPE, REMOVE_POPUP);
	if (ret < 0)
		_E("Failed to launch remove battery popup(%d)", ret);
}

static int check_power_supply_noti(void)
{
	return 1;
}

static void update_ovp(enum battery_noti_status status)
{
	_I("charge %d ovp %d(old %d)", battery->charge_now, battery->health, old_battery.health);

	old_battery.health = battery->health;
	if (status == DEVICE_NOTI_ON)
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
	else
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	device_notify(DEVICE_NOTIFIER_BATTERY_OVP, (void *)&battery->health);
}

static void health_timer_reset(void)
{
	abnormal_timer = 0;
}

static gboolean health_timer_cb(void *data)
{
	health_timer_reset();

	if (battery->health != HEALTH_LOW && battery->health != HEALTH_HIGH)
		return G_SOURCE_REMOVE;

	CRITICAL_LOG("Popup: Battery health status is not good, %s.", battery->health_s);
	device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery->health);
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
	if (disp_plgn->pm_unlock_internal)
		disp_plgn->pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
	if (disp_plgn->pm_lock_internal)
		disp_plgn->pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	if (battery->health == HEALTH_LOW)
		battery_charge_err_low_act(NULL);
	else if (battery->health == HEALTH_HIGH)
		battery_charge_err_high_act(NULL);
	return G_SOURCE_REMOVE;
}

static void abnormal_popup_dbus_signal_handler(GDBusConnection  *conn,
		const gchar      *sender,
		const gchar      *path,
		const gchar      *iface,
		const gchar      *name,
		GVariant         *param,
		gpointer          user_data)

{
	if (battery->health == HEALTH_GOOD)
		return;

	_I("Restart health timer.");
	abnormal_timer = g_timeout_add_seconds(ABNORMAL_CHECK_TIMER_INTERVAL,
						health_timer_cb, NULL);
	if (abnormal_timer == 0)
		_E("Failed to add abnormal check timer.");
}

static int display_changed(void *data)
{
	static enum state_t state;

	if (!data)
		return state;

	state = *(int *)data;

	return state;
}

static void battery_notification_init(void *data)
{
	struct battery_plugin *plugin = (struct battery_plugin *)data;
	int ret;

	if (!plugin)
		return;

	_D("Add plugins for battery notification.");
	plugin->lowbat_noti_launch = launch_lowbat_noti;
	plugin->lowbat_noti_clean = clean_lowbat_noti;

	plugin->changed_battery_cf = changed_battery_cf;
	plugin->remove_health_popup = remove_health_popup;
	plugin->check_power_supply_noti = check_power_supply_noti;
	plugin->update_ovp = update_ovp;

	plugin->display_changed = display_changed;

	ret = subscribe_dbus_signal(NULL, DEVICED_PATH_SYSNOTI,
		DEVICED_INTERFACE_SYSNOTI, SIGNAL_CHARGEERR_RESPONSE, abnormal_popup_dbus_signal_handler, NULL, NULL);
	if (ret <= 0)
		_E("Failed to init dbus signal: %d", ret);

	register_notifier(DEVICE_NOTIFIER_LCD, display_changed);
}

static const struct battery_ops battery_notification_ops = {
	.name     = "battery_notification",
	.init     = battery_notification_init,
};

BATTERY_OPS_REGISTER(&battery_notification_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_display_plugin();
	if (!disp_plgn)
		_E("Failed to get display plugin.");

	battery = get_battery_status();
	if (!battery)
		_E("Failed to get battery status structure.");
}
