/*
 * deviced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <vconf.h>

#include "core/common.h"
#include "core/devices.h"
#include "core/device-notifier.h"
#include "core/log.h"
#include "display/core.h"
#include "display/device-interface.h"
#include "display/display-ops.h"
#include "display-info.h"
#include "powersaver.h"
#include "setting.h"

static struct _backlight_ops *backlight_ops;

static int set_powersaver_mode(int mode)
{
	int timeout;
	const struct display_ops *hbm_ops = NULL;
	struct hbmsetstate hss;
	bool pmqos = false;

	if (mode == POWERSAVER_ENHANCED)
		pmqos = true;
	_D("Powersaver mode %d pmqos %d.", mode, pmqos);
	device_notify(DEVICE_NOTIFIER_ULTRAPOWERSAVING, (void *)&pmqos);

	backlight_ops->set_force_brightness(0);
	set_force_lcdtimeout(0);

	FIND_DISPLAY(hbm_ops, "hbm");
	if (hbm_ops && hbm_ops->func(HBM_GET_STATE, 0)) {
		hss.hbm = false;
		hss.timeout = 0;
		hbm_ops->func(HBM_SET_TIMEOUT_STATE, &hss);
	}

	backlight_ops->update();
	get_run_timeout(&timeout);
	state_st(S_NORMAL)->timeout = timeout;
	state_st(get_pm_cur_state())->trans(EVENT_INPUT);

	return 0;
}

static void powersaver_status_changed(keynode_t *key_nodes, void *data)
{
	int status, mode, ret;

	if (key_nodes == NULL) {
		_E("Wrong parameter, key_nodes is null.");
		return;
	}

	status = vconf_keynode_get_int(key_nodes);
	if (status == SETTING_PSMODE_NORMAL)
		mode = POWERSAVER_OFF;
	else if (status == SETTING_PSMODE_WEARABLE)
		mode = POWERSAVER_BASIC;
	else if (status == SETTING_PSMODE_WEARABLE_ENHANCED)
		mode = POWERSAVER_ENHANCED;
	else
		return;

	ret = set_powersaver_mode(mode);
	if (ret < 0)
		_E("Failed to update powersaver state %d.", ret);
}

static int booting_done(void *data)
{
	static int done;
	int ret, status;
	int mode = POWERSAVER_OFF;
	bool pmqos = false;

	if (data == NULL)
		goto out;
	done = *(int *)data;
	if (!done)
		goto out;
	_I("Booting done.");


	ret = vconf_notify_key_changed(VCONFKEY_SETAPPL_PSMODE,
	    powersaver_status_changed, NULL);
	if (ret != 0)
		_E("Failed to vconf_notify_key_changed.");

	ret = vconf_get_int(VCONFKEY_SETAPPL_PSMODE, &status);
	if (ret != 0) {
		_E("Failed to vconf get bool.");
		goto out;
	}

	switch (status) {
	case SETTING_PSMODE_WEARABLE:
		mode = POWERSAVER_BASIC;
		break;
	case SETTING_PSMODE_WEARABLE_ENHANCED:
		mode = POWERSAVER_ENHANCED;
		pmqos = true;
		break;
	default:
		goto out;
	}

	_D("Powersaver mode on, %d.", mode);
	device_notify(DEVICE_NOTIFIER_PMQOS_ULTRAPOWERSAVING, (void *)&pmqos);
	ret = set_powersaver_mode(mode);
	if (ret < 0)
		_E("Failed to update powersaver state, %d.", ret);
out:
	return done;
}

static void powersaver_init(void *data)
{
	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void powersaver_exit(void *data)
{
	int ret;

	ret = vconf_ignore_key_changed(VCONFKEY_SETAPPL_PSMODE,
	    powersaver_status_changed);
	if (ret != 0)
		_E("Failed to vconf_ignore_key_changed.");
}

static const struct device_ops powersaver_device_ops = {
	DECLARE_NAME_LEN("powersaver"),
	.init     = powersaver_init,
	.exit     = powersaver_exit,
};

DEVICE_OPS_REGISTER(&powersaver_device_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	backlight_ops = get_backlight_ops();
	if (!backlight_ops)
		_E("Failed to get backlight operator.");
}
