/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>

#include "core/device-notifier.h"
#include "core/devices.h"
#include "display/core.h"
#include "display/util.h"
#include "display/display-ops.h"
#include "shared/plugin.h"

#include <libsyscommon/dbus-system.h>

#define DBUS_COREAPPS_SWIMMODE_PATH		"/Org/Tizen/Coreapps/home/swimmode"
#define DBUS_COREAPPS_SWIMMODE_INTERFACE	"org.tizen.coreapps.home.swimmode"
#define DBUS_SWIMMODE_MEMBER_STATUS		"status"

static struct display_plugin *disp_plgn;
static struct _backlight_ops *backlight_ops;
static const struct device_ops *touchscreen_ops;
static int swimmode_status;

static int swim_execute(void *data)
{
	int state;

	if (check_default(touchscreen_ops))
		return 0;

	state = (intptr_t) data;
	if (swimmode_status == state)
		return 0;

	swimmode_status = state;

	/**
	 * When it is enable, TSP have to change to off and it must not control.
	 * So if this signal is received, tsp mode would be stopped.
	 * After that, module would be changed to suspend(execute(STATUS_STOP)).
	 * The opposite case, the module should be enabled first.
	 */
	if (swimmode_status) {
		touchscreen_ops->stop(TOUCH_SCREEN_OFF_MODE);
		touchscreen_ops->execute((void *)DEVICE_OPS_STATUS_STOP);
	} else {
		touchscreen_ops->execute((void *)DEVICE_OPS_STATUS_START);
		touchscreen_ops->start(NORMAL_MODE);
	}

	return 0;
}

static void swimmode_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int val = 0;
	pid_t pid;
	bool lcd_state = false;

	if (!dh_get_param_from_var(param, "(i)", &val)) {
		_E("failed to get params from gvariant. expected:%s, type:%s", "(i)", g_variant_get_type_string(param));
		return;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (pid == -1 || kill(pid, 0) == -1) {
		_E("Sender(%d) does not exist.", pid);
		return;
	}

	_I("Request swim mode state %d by %d", val, pid);

	if (backlight_ops->get_lcd_power != DPMS_ON) {
		if (disp_plgn->pm_change_internal)
			disp_plgn->pm_change_internal(INTERNAL_LOCK_SWIM, S_NORMAL);
		lcd_state = true;
	}

	swim_execute((void *)((intptr_t)val));

	if (lcd_state) {
		if (disp_plgn->pm_change_internal)
			disp_plgn->pm_change_internal(INTERNAL_LOCK_SWIM, S_LCDOFF);
	}
}

static int swim_restore_cb(void *data)
{
	_I("Swim mode is restored.");
	swim_execute((void *)0);

	return 0;
}

static void swim_init(void *data)
{
	int ret;

	touchscreen_ops = find_device("touchscreen");

	ret = subscribe_dbus_signal(NULL,
				DBUS_COREAPPS_SWIMMODE_PATH,
				DBUS_COREAPPS_SWIMMODE_INTERFACE,
				DBUS_SWIMMODE_MEMBER_STATUS,
				swimmode_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	register_notifier(DEVICE_NOTIFIER_LONGKEY_RESTORE, swim_restore_cb);
}

static const struct device_ops swim_device_ops = {
	DECLARE_NAME_LEN("swim"),
	.init     = swim_init,
	.execute  = swim_execute,
};

DEVICE_OPS_REGISTER(&swim_device_ops)

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_display_plugin();
	if (!disp_plgn) {
		_E("Failed to get display plugin.");
	}

	backlight_ops = get_backlight_ops();
	if (!backlight_ops)
		_E("Failed to get backlight operator.");
}
