/*
 * deviced
 *
 * Copyright (c) 2012 - 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdarg.h>
#include "core/log.h"
#include "core/common.h"
#include "apps.h"
#include "shared/plugin.h"

#define POPUP_METHOD "PopupLaunch"
#define BUFF_MAX        255

static struct display_plugin *disp_plgn;

static const struct app_dbus_match {
	const char *type;
	const char *bus;
	const char *path;
	const char *iface;
	const char *method;
} app_match[] = {
	{ APP_DEFAULT , POPUP_BUS_NAME, POPUP_PATH_SYSTEM  , POPUP_INTERFACE_SYSTEM  , POPUP_METHOD },
	{ APP_POWERKEY, POPUP_BUS_NAME, POPUP_PATH_POWERKEY, POPUP_INTERFACE_POWERKEY, POPUP_METHOD },
	{ APP_OVERHEAT, POPUP_BUS_NAME, POPUP_PATH_OVERHEAT, POPUP_INTERFACE_OVERHEAT, POPUP_METHOD },
	{ APP_ABNORMAL, POPUP_BUS_NAME, POPUP_PATH_SYSTEM  , POPUP_INTERFACE_SYSTEM  , POPUP_METHOD },
	{ APP_REMOVE  , POPUP_BUS_NAME, POPUP_PATH_SYSTEM  , POPUP_INTERFACE_SYSTEM  , POPUP_METHOD },
};

static void __cb(GVariant *var, void *user_data, GError *err)
{
	 int ret;

	if (!var) {
		_E("No message: %s", err->message);
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &ret)) {
		_E("No message: %s", g_variant_get_type_string(var));
		goto out;
	}

	_D("Reply value: %d", ret);

out:
	g_variant_unref(var);
}


int launch_system_app(char *type, int num, ...)
{
	char *app_type;
	va_list args;
	int i, match, ret;

	if (type)
		app_type = type;
	else
		app_type = APP_DEFAULT;

	match = -1;
	for (i = 0 ; i < ARRAY_SIZE(app_match) ; i++) {
		if (strncmp(app_type, app_match[i].type, strlen(app_type)))
		continue;
		match = i;
		break;
	}
	if (match < 0) {
		_E("Failed to find matched app type(%s).", app_type);
		return -EINVAL;
	}

	va_start(args, num);

	ret = dbus_handle_method_async_pairs_with_reply(app_match[match].bus,
		app_match[match].path,
		app_match[match].iface,
		app_match[match].method,
		num,
		args,
		__cb,
		-1,
		NULL);

	va_end(args);

	if (disp_plgn->pm_change_internal)
		disp_plgn->pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);

	return ret;
}

int launch_message_post(char *type)
{
	int ret;

	if (!type)
		return -EINVAL;

	ret = dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
			POPUP_PATH_NOTI,
			POPUP_INTERFACE_NOTI,
			"MessagePostOn",
			g_variant_new("(s)", type),
			__cb, -1, NULL);

	if (disp_plgn->pm_change_internal)
		disp_plgn->pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	return ret;
}

int add_async_notification(char *type, dbus_pending_cb func, char *sig, ...)
{
	int ret;
	int count;
	int len;
	char *data = NULL;
	const char *param[BUFF_MAX] = {NULL};
	va_list args;

	if (!type)
		return -EINVAL;

	if (sig) {
		len = strlen(sig);
		va_start(args, sig);

		for (count = 0; count < len; count++) {
			data = va_arg(args, char *);
			param[count] = data;
		}

		va_end(args);
	}

	ret = dbus_handle_method_async_with_reply(POPUP_BUS_NAME,
			POPUP_PATH_NOTI,
			POPUP_INTERFACE_NOTI,
			type,
			sig,
			param,
			func,
			-1,
			NULL);

	return ret;
}

int remove_notification(char *type, int id)
{
	if (!type || id < 0)
		return -EINVAL;

	return dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
			POPUP_PATH_NOTI,
			POPUP_INTERFACE_NOTI,
			type, g_variant_new("(i)", id),
			__cb, -1, NULL);
}

static void __CONSTRUCTOR__ initialize(void)
{
	disp_plgn = get_display_plugin();
	if (!disp_plgn) {
		_E("Failed to get display plugin.");
	}
}
