/*
 * test
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <error.h>
#include <tzplatform_config.h>
#include <device/display-internal.h>

#include "test.h"

#define METHOD_DISPLAY_GETBRIGHTNESS		"getbrightness"
#define METHOD_DISPLAY_SETBRIGHTNESS		"setbrightness"
#define METHOD_DISPLAY_GETMAXBRIGHTNESS		"GetMaxBrightness"
#define METHOD_DISPLAY_SETMAXBRIGHTNESS		"SetMaxBrightness"
#define METHOD_DISPLAY_HOLDBRIGHTNESS		"HoldBrightness"
#define METHOD_DISPLAY_RELEASEBRIGHTNESS	"ReleaseBrightness"
#define METHOD_DISPLAY_CUSTOMBRIGHTNESS		"CustomBrightness"
#define METHOD_DISPLAY_SETAUTOBRIGHTNESSMIN	"setautobrightnessmin"
#define METHOD_DISPLAY_SETREFRESHRATE		"SetRefreshRate"
#define METHOD_DISPLAY_GETDISPLAYCOUNT		"GetDisplayCount"
#define METHOD_DISPLAY_CHANGESTATE			"ChangeState"
#define METHOD_DISPLAY_LOCKSTATE_NOPT		"lockstate_nopt"
#define METHOD_DISPLAY_UNLOCKSTATE			"unlockstate"
#define METHOD_DISPLAY_SETLCDTIMEOUT		"setlcdtimeout"
#define METHOD_DISPLAY_CUSTOMLCDON			"CustomLCDOn"
#define METHOD_DISPLAY_LCDPANEL_OFFMODE		"LCDPanelOffMode"
#define METHOD_DISPLAY_STAYTOUCHSCREEN_OFF	"StayTouchScreenOff"
#define METHOD_DISPLAY_LOCKSCREENBGON		"LockScreenBgOn"
#define METHOD_DISPLAY_DUMPMODE				"Dumpmode"
#define METHOD_DISPLAY_SAVELOG				"SaveLog"
#define METHOD_DISPLAY_POWERKEY_IGNORE		"PowerKeyIgnore"
#define METHOD_DISPLAY_POWERKEY_LCDOFF		"PowerKeyLCDOff"
#define METHOD_DISPLAY_LOCKTIMEOUT_EXPIRED	"LockTimeoutExpired"
#define METHOD_DISPLAY_LOCKTIMEOUT_INPUT	"LockTimeoutInput"
#define METHOD_DISPLAY_ACTORCONTROL			"ActorControl"
#define METHOD_DISPLAY_START				"start"
#define METHOD_DISPLAY_STOP					"stop"
#define METHOD_CHANGE_STATE_BY_REASON		"ChangeStateByReason"

#define PM_STATE_LOG_FILE			"/var/log/pm_state.log"
#define PALM_STR				"palm"

static bool get_display_method(const char *method, GVariant *param, int *value)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			method, param);
	if (!msg) {
		_E("fail (%s): no reply", method);
		return ret;
	}
	if (!dh_get_param_from_var(msg, "(i)", &val))
		_E("fail (%s): no message", method);
	else {
		_I("success (%s): %d", method, val);
		ret = TRUE;
		if (NULL != value)
			*value = val;
	}

	g_variant_unref(msg);
	return ret;
}

static bool set_display_method(const char *method, GVariant *param)
{
	GVariant *msg;
	int val;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			method, param);
	if (!msg) {
		_E("fail (%s): no reply", method);
		return ret;
	}

	if (!dh_get_param_from_var(msg, "(i)", &val))
		_E("fail (%s): no message", method);
	else {
		if ((val == -ENOTSUP) || (val == -ENOSYS)) {
			_I("Not supported feature! (%s): %d", method, val);
			ret = TRUE;
		} else if (val < 0) {
			_E("fail (%s): returned fail (%d)", method, val);
		} else {
			_I("success (%s): %d", method, val);
			ret = TRUE;
		}
	}

	g_variant_unref(msg);
	return ret;
}

static bool set_display_method_noreply(const char *method, GVariant *var)
{
	GVariant *msg;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			method, var);
	if (!msg) {
		_E("fail (%s): no reply", method);
		return FALSE;
	}

	g_variant_unref(msg);
	return TRUE;
}

static bool set_display_maxbrightness(int brightness)
{
	return set_display_method(METHOD_DISPLAY_SETMAXBRIGHTNESS, g_variant_new("(i)", brightness));
}

static bool set_display_holdbrightness(int brightness)
{
	return set_display_method(METHOD_DISPLAY_HOLDBRIGHTNESS, g_variant_new("(i)", brightness));
}

static bool set_display_releasebrightness()
{
	return set_display_method(METHOD_DISPLAY_RELEASEBRIGHTNESS, NULL);
}

static bool get_display_custombrightness()
{
	return get_display_method(METHOD_DISPLAY_CUSTOMBRIGHTNESS, NULL, NULL);
}

static bool set_display_autobrightness_min(int min)
{
	GVariant *msg;
	int val = 0;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						METHOD_DISPLAY_SETAUTOBRIGHTNESSMIN,
						g_variant_new("(i)", min));
	if (!msg) {
		_E("fail : no reply");
		return ret;
	}

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("fail : no message");
	} else {
		if ((min < 1) || (min > 100)) {
			if (val < 0) {
				_I("success : autobrightness min value should be 1~100. can't set (%d).", min);
				ret = TRUE;
			} else {
				_I("fail : can't set (%d). but returned success", min);
			}
		} else {
			if (val < 0) {
				_E("fail : return fail (%d)", val);
			} else {
				_I("success : %d", val);
				ret = TRUE;
			}
		}
	}
	g_variant_unref(msg);
	return ret;
}

#if 0
static bool set_display_refreshrate(int app, int rate)
{
	return set_display_method(METHOD_DISPLAY_SETREFRESHRATE, g_variant_new("(ii)", app, rate));
}
#endif

static bool get_display_count()
{
	return get_display_method(METHOD_DISPLAY_GETDISPLAYCOUNT, NULL, NULL);
}

static bool set_display_changestate(char *state)
{
	return set_display_method(METHOD_DISPLAY_CHANGESTATE, g_variant_new("(s)", state));
}

#if 0
static bool set_display_lockstate_nopt(char *state, char *option, char *ignore, int timeout)
{
	char *param[4];
	char str_timeout[10];

	snprintf(str_timeout, sizeof(str_timeout), "%d", timeout);

	param[0] = state;
	param[1] = option;
	param[1] = ignore;
	param[1] = str_timeout;

	return set_display_method(METHOD_DISPLAY_LOCKSTATE_NOPT, "sssi", param);
}

static bool set_display_unlockstate(char *state, char *option)
{
	char *param[2];

	param[0] = state;
	param[1] = option;

	return set_display_method(METHOD_DISPLAY_UNLOCKSTATE, "ss", param);
}
#endif

static bool set_display_lcdtimeout(int normal, int dim, int holdkey_block)
{
	return set_display_method(METHOD_DISPLAY_SETLCDTIMEOUT, g_variant_new("(iii)", normal, dim, holdkey_block));
}

static bool set_display_customlcdon(int timeout_msec)
{
	return set_display_method(METHOD_DISPLAY_CUSTOMLCDON, g_variant_new("(i)", timeout_msec));
}

static bool set_display_lcdpanel_offmode(int on)
{
	return set_display_method(METHOD_DISPLAY_LCDPANEL_OFFMODE, g_variant_new("(i)", on));
}

static bool set_display_staytouchscreen_off(int on)
{
	return set_display_method(METHOD_DISPLAY_STAYTOUCHSCREEN_OFF, g_variant_new("(i)", on));
}

static bool set_display_lockscreen_bg_on(char *on)
{
	return set_display_method(METHOD_DISPLAY_LOCKSCREENBGON, g_variant_new("(s)", on));
}

static bool set_display_dumpmode(char *on)
{
	return set_display_method(METHOD_DISPLAY_DUMPMODE, g_variant_new("(s)", on));
}

static bool set_display_savelog()
{
	struct stat buf;
	int r;
	bool ret = FALSE;

	if (!stat(PM_STATE_LOG_FILE, &buf)) {
		r = remove(PM_STATE_LOG_FILE);
		if (r < 0)
			_E("Remove file error : %d", r);
	}

	if (!set_display_method_noreply(METHOD_DISPLAY_SAVELOG, NULL))
		return ret;

	sleep(3);

	if (!stat(PM_STATE_LOG_FILE, &buf)) {
		_I("success : %s", PM_STATE_LOG_FILE);
		ret = TRUE;
	}

	return ret;
}

static bool set_display_powerkey_ignore(int on)
{
	return set_display_method(METHOD_DISPLAY_POWERKEY_IGNORE, g_variant_new("(i)", on));
}

static bool set_display_powerkey_lcdoff()
{
	return set_display_method(METHOD_DISPLAY_POWERKEY_LCDOFF, NULL);
}

#if 0
static bool set_display_locktimeout_expired(char *req_id)
{
	DBusMessage *msg;
	int val;
	char *ret_req_id;
	bool ret = FALSE;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			METHOD_DISPLAY_LOCKTIMEOUT_EXPIRED, g_variant_new("(s)", req_id));
	if (!msg) {
		_E("fail: no reply");
		return ret;
	}

	if (dbus_message_get_args(msg, NULL,
				DBUS_TYPE_STRING, &ret_req_id, //request id
				DBUS_TYPE_INT32, &val,   //return value
				DBUS_TYPE_INVALID) == 0)
		_E("fail: no message");
	else {
		if ((0 != strcmp(req_id, ret_req_id)) || (val < 0)) {
			_E("fail: return fail");
		} else {
			_I("success: %s %d", ret_req_id, val);
			ret = TRUE;
		}
	}

	dbus_message_unref(msg);
	return ret;
}

static bool set_display_locktimeout_input(char *req_id, int input)
{
	char *param[2];
	char str_input[10];

	snprintf(str_input, sizeof(str_input), "%d", input);
	param[0] = req_id;
	param[1] = str_input;

	return set_display_method(METHOD_DISPLAY_LOCKTIMEOUT_INPUT, g_variant_new("(si)", req_id, input));
}
#endif

static bool set_display_actorcontrol(char *command, int actor, int caps)
{
	return set_display_method(METHOD_DISPLAY_ACTORCONTROL, g_variant_new("(sii)", command, actor, caps));
}

static bool get_display_brightness_state(display_state_e state, int *brightness)
{
	int ret;

	ret = device_display_get_brightness_state(0, state, brightness);

	return capi_result(METHOD_DISPLAY_GETBRIGHTNESS, ret);
}

static bool get_display_max_brightness_state(display_state_e state, int *brightness)
{
	int ret;

	ret = device_display_get_max_brightness_state(0, state, brightness);

	return capi_result(METHOD_DISPLAY_GETMAXBRIGHTNESS, ret);
}

static bool set_display_brightness_state(display_state_e state, int brightness)
{
	int ret;
	int getbrightness;

	ret = device_display_set_brightness_state(0, state, brightness);
	if (!capi_result(METHOD_DISPLAY_SETBRIGHTNESS, ret))
		return FALSE;

	if (state == DISPLAY_STATE_NORMAL)
		set_display_changestate("lcdon");
	else if (state == DISPLAY_STATE_SCREEN_DIM)
		set_display_changestate("lcddim");
	else {
		_E("fail: unknown state(%d)", state);
		return FALSE;
	}

	ret = device_display_get_brightness_state(0, state, &getbrightness);
	if (!capi_result(METHOD_DISPLAY_SETBRIGHTNESS, ret))
		return FALSE;

	if (brightness != getbrightness)
		return FALSE;

	return TRUE;
}

static bool set_display_state_by_reason(display_state_e type, const char *reason, int timeout, device_dbus_pending_cb cb)
{
	int ret;

	ret = device_display_change_state_by_reason(type, reason, timeout, cb);

	return capi_result(METHOD_CHANGE_STATE_BY_REASON, ret);
}

static bool set_display_start()
{
	return set_display_method_noreply(METHOD_DISPLAY_START, NULL);
}

static bool set_display_stop()
{
	return set_display_method_noreply(METHOD_DISPLAY_STOP, NULL);
}

void display_test_all(int *success, int *fail)
{
	int s = 0;
	int f = 0;
	int brt = 0;

	(get_display_brightness_state(DISPLAY_STATE_NORMAL, &brt))	       ? s++ : f++;
	(get_display_brightness_state(DISPLAY_STATE_SCREEN_DIM, &brt))	   ? s++ : f++;
	(set_display_brightness_state(DISPLAY_STATE_NORMAL, 100))          ? s++ : f++;
	(set_display_brightness_state(DISPLAY_STATE_SCREEN_DIM, 1))        ? s++ : f++;
	(get_display_max_brightness_state(DISPLAY_STATE_NORMAL, &brt))				? s++ : f++;
	(get_display_max_brightness_state(DISPLAY_STATE_SCREEN_DIM, &brt))			? s++ : f++;
	(set_display_maxbrightness(80))		? s++ : f++;	//not support, always success.
	(set_display_holdbrightness(30))	? s++ : f++;
	(set_display_releasebrightness())	? s++ : f++;
	(get_display_custombrightness())	? s++ : f++;
	(set_display_autobrightness_min(15)) ? s++ : f++;
	(set_display_autobrightness_min(0))	? s++ : f++;
	//(set_display_refreshrate(0, 30))	? s++ : f++;
	(get_display_count())				? s++ : f++;
	(set_display_changestate("lcdon"))	? s++ : f++;
	//(set_display_lockstate_nopt("lcddim", "staycurstate", "", 3)) ? s++ : f++;
	//(set_display_unlockstate("lcddim", "resettimer")) ? s++ : f++;
	(set_display_lcdtimeout(5, 10, 1))	? s++ : f++;
	(set_display_customlcdon(6000))		? s++ : f++;
	(set_display_lcdpanel_offmode(1))	? s++ : f++;
	(set_display_lcdpanel_offmode(0))	? s++ : f++;
	(set_display_staytouchscreen_off(1)) ? s++ : f++;
	(set_display_staytouchscreen_off(0)) ? s++ : f++;
	(set_display_lockscreen_bg_on("true"))	? s++ : f++;
	(set_display_dumpmode("on"))		? s++ : f++;
	(set_display_dumpmode("off"))		? s++ : f++;
	(set_display_savelog())				? s++ : f++;
	(set_display_powerkey_ignore(1))	? s++ : f++;	//not support, always success.
	(set_display_powerkey_lcdoff())		? s++ : f++;	//not support, always success.
	//(set_display_locktimeout_expired("auto_test")) ? s++ : f++;
	//(set_display_locktimeout_input("auto_test", 1)) ? s++ : f++;
	(set_display_actorcontrol("set", 4, 1<<3))   ? s++ : f++;
	(set_display_actorcontrol("reset", 4, 1<<3)) ? s++ : f++;
	(set_display_stop())				? s++ : f++;
	(set_display_start())				? s++ : f++;
	(set_display_state_by_reason(DISPLAY_STATE_SCREEN_OFF, PALM_STR, 0, __cb)) ? s++ : f++;

	if (NULL != success)    *success = s;
	if (NULL != fail)       *fail = f;
}

static void display_init(void *data)
{
	int success = 0;
	int fail = 0;

	_I("start test");

	display_test_all(&success, &fail);

	_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
}

static void display_exit(void *data)
{
	_I("end test");

}

static int display_unit(int argc, char **argv)
{
	int brt = 0;

	if (argc < 4) {
		int success = 0;
		int fail = 0;
		_I("start test");
		display_test_all(&success, &fail);
		_I("Total: %d, Success: %d, Fail: %d", success+fail, success, fail);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_GETBRIGHTNESS)) {
		get_display_brightness_state(DISPLAY_STATE_NORMAL, &brt);
		get_display_brightness_state(DISPLAY_STATE_SCREEN_DIM, &brt);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SETBRIGHTNESS)) {
		set_display_brightness_state(DISPLAY_STATE_NORMAL, 100);
		set_display_brightness_state(DISPLAY_STATE_SCREEN_DIM, 1);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_GETMAXBRIGHTNESS)) {
		get_display_max_brightness_state(DISPLAY_STATE_NORMAL, &brt);
		get_display_max_brightness_state(DISPLAY_STATE_SCREEN_DIM, &brt);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SETMAXBRIGHTNESS)) {
		set_display_maxbrightness(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_HOLDBRIGHTNESS)) {
		set_display_holdbrightness(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_RELEASEBRIGHTNESS)) {
		set_display_releasebrightness();
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_CUSTOMBRIGHTNESS)) {
		get_display_custombrightness();
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SETAUTOBRIGHTNESSMIN)) {
		set_display_autobrightness_min(atoi(argv[4]));
//	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SETREFRESHRATE)) {
//		set_display_refreshrate(atoi(argv[4]), atoi(argv[5]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_GETDISPLAYCOUNT)) {
		get_display_count();
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_CHANGESTATE)) {
		set_display_changestate(argv[4]);
//	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_LOCKSTATE_NOPT)) {
//		set_display_lockstate_nopt(argv[4], argv[5], argv[6], atoi(argv[7]));
//	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_UNLOCKSTATE)) {
//		set_display_lockstate_nopt(argv[4], argv[5]);
	} else if (0 == strcasecmp(argv[3], METHOD_CHANGE_STATE_BY_REASON)) {
		set_display_state_by_reason(DISPLAY_STATE_SCREEN_OFF, PALM_STR, 0, __cb);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SETLCDTIMEOUT)) {
		set_display_lcdtimeout(atoi(argv[4]), atoi(argv[5]), atoi(argv[6]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_CUSTOMLCDON)) {
		set_display_customlcdon(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_LCDPANEL_OFFMODE)) {
		set_display_lcdpanel_offmode(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_STAYTOUCHSCREEN_OFF)) {
		set_display_staytouchscreen_off(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_LOCKSCREENBGON)) {
		set_display_lockscreen_bg_on(argv[4]);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_DUMPMODE)) {
		set_display_dumpmode(argv[4]);
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_SAVELOG)) {
		set_display_savelog();
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_POWERKEY_IGNORE)) {
		set_display_powerkey_ignore(atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_POWERKEY_LCDOFF)) {
		set_display_powerkey_lcdoff();
//	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_LOCKTIMEOUT_EXPIRED)) {
//		set_display_locktimeout_expired(argv[4]);
//	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_LOCKTIMEOUT_INPUT)) {
//		set_display_locktimeout_input(argv[4], atoi(argv[4]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_ACTORCONTROL)) {
		set_display_actorcontrol(argv[4], atoi(argv[5]), atoi(argv[6]));
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_START)) {
		set_display_start();
	} else if (0 == strcasecmp(argv[3], METHOD_DISPLAY_STOP)) {
		set_display_stop();
	} else {
		_E("Unknown test case!!!");
	}

	return 0;
}

static const struct test_ops display_test_ops = {
	.priority = TEST_PRIORITY_NORMAL,
	.name     = "display",
	.init     = display_init,
	.exit    = display_exit,
	.unit    = display_unit,
};

TEST_OPS_REGISTER(&display_test_ops)
