/*
 * deviced
 *
 * Copyright (c) 2019 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <hw/board.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/devices.h"

static struct hw_board *board_dev;

static GVariant *dbus_revision_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	int revision;

	if (!board_dev || !board_dev->get_device_revision) {
		_E("GetHWRev not supported.");
		ret = -ENOTSUP;
		goto revision_fail;
	}

	ret = board_dev->get_device_revision(&revision);
	if (ret < 0) {
		_E("Fail to get revision.");
		goto revision_fail;
	}

	_D("Revision: %d", revision);

	return g_variant_new("(i)", revision);

revision_fail:
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_num_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	GVariant *result;
	char *serial = NULL;

	if (!board_dev || !board_dev->get_device_serial) {
		_E("GetSerial not supported.");
		ret = -ENOTSUP;
		goto num_out;
	}

	ret = board_dev->get_device_serial(&serial);
	if (ret < 0) {
		_E("Failed to get serial.");
		goto num_out;
	}

	_D("Num(%s) len(%zu).", serial, strlen(serial));

num_out:
	if (serial == NULL)
		serial = strdup("");

	result = g_variant_new("(si)", serial, ret);

	if (serial)
		free(serial);

	return result;
}

static GVariant *dbus_serial_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	char *p;
	GVariant *result;
	char *num = NULL;

	if (!board_dev || !board_dev->get_device_serial) {
		_E("GetNum not supported.");
		ret = -ENOTSUP;
		goto seiral_out;
	}

	ret = board_dev->get_device_serial(&num);
	if (ret < 0) {
		_E("Failed to get num.");
		goto seiral_out;
	}

	p = strchr(num, ',');
	if (p)
		*p = '\0';

	_D("Serial(%s) len(%zu).", num, strlen(num));

seiral_out:
	if (num == NULL)
		num = strdup("");

	result = g_variant_new("(si)", num, ret);

	if (num)
		free(num);

	return result;
}

static const dbus_method_s dbus_methods[] = {
	{ "GetSerial", 	NULL, "si",	dbus_serial_handler },
	{ "GetHWRev",	NULL, "i",	dbus_revision_handler },
	{ "GetNum",		NULL, "si",	dbus_num_handler},
};

static const dbus_interface_u dbus_interface = {
	.oh 		= NULL,
	.name 		= DEVICED_INTERFACE_BOARD,
	.methods 	= dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int board_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (board_dev)
		return 0;

	ret = hw_get_info(BOARD_HARDWARE_DEVICE_ID,
			(const struct hw_info**)&info);
	if (ret < 0) {
		_E("Failed to load board shared library: %d", ret);
		return -ENODEV;
	}

	if (!info->open) {
		_E("Failed to open board device: open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&board_dev);
	if (ret < 0) {
		_E("Failed to get board device structure: %d", ret);
		return -ENODEV;
	}

	_I("Board device structure load success.");
	return 0;
}

static void board_init(void *data)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_BOARD, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static const struct device_ops board_device_ops = {
	DECLARE_NAME_LEN("board"),
	.probe	= board_probe,
	.init	= board_init,
};

DEVICE_OPS_REGISTER(&board_device_ops)
