/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <fcntl.h>
#include <sys/reboot.h>
#include <systemd/sd-daemon.h>
#include <glib.h>
#include <libsyscommon/dbus-system.h>
#include <libsyscommon/systemd-state.h>
#include <device/booting-internal.h>
#include <argos.h>

#include "display/core.h"
#include "log.h"
#include "common.h"
#include "devices.h"
#include "power/boot.h"
#include "power/power-handler.h"
#include "shared/plugin.h"
#include "device-notifier.h"

#define PIDFILE_PATH		"/var/run/.deviced.pid"
#define WATCHDOG_REFRESH_TIME	5
#define WATCHDOG_TIMEOUT	90

static GMainLoop *mainloop;
int silent_boot;

static void writepid(char *pidpath)
{
	FILE *fp;

	fp = fopen(pidpath, "w");
	if (fp != NULL) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}
}

static void sig_quit(int signo)
{
	_D("Received SIGTERM signal(%d).", signo);
}

static void sig_usr1(int signo)
{
	CRITICAL_LOG("Received SIGUSR1 signal(%d), deviced'll be finished.", signo);
	if (mainloop && g_main_loop_is_running(mainloop))
		g_main_loop_quit(mainloop);
}

void watchdog_notify(void)
{
	int ret = aw_notify();
	if (ret < 0)
		_E("Failed to aw_notifty: %d", ret);
}

static void deviced_dbus_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
	int ret;
	ret = check_system_boot_finished();
	if (ret == 1) {
		/* Restarted: deviced was terminated */
		_I("Notify relaunch.");
		device_notify_once(DEVICE_NOTIFIER_BOOTING_DONE, &ret);
	}

	_I("sd_notify(READY=1)");
	sd_notify(0, "READY=1");
}

static gboolean watchdog_cb(void *data)
{
	watchdog_notify();
	return G_SOURCE_CONTINUE;
}

static int deviced_main(int argc, char **argv)
{
	int ret;
	dbus_handle_h handle = NULL;
	guint timer;

	CRITICAL_LOG("Initializing deviced.");
	mainloop = g_main_loop_new(NULL, FALSE);

	ret = check_power_flag();
	if (ret) {
		/* Restarted: deviced was terminated
		 * in middle of reboot/poweroff - resume procedure
		 */
		poweroff_request_shutdown();
		return 0;
	}

	handle = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, FALSE);
	if (!handle)
		_E("Failed to get dbus connection.");

	load_plugins();

	silent_boot = device_get_reboot_mode();

	devices_init(NULL);

	ret = dbus_handle_request_bus_name(handle, DEVICED_BUS_NAME, deviced_dbus_name_acquired, NULL);
	if (ret <= 0) {
		_E("Failed to request bus name.");
		dbus_handle_check_owner_name(NULL, DEVICED_BUS_NAME);
	}

	signal(SIGTERM, sig_quit);
	signal(SIGUSR1, sig_usr1);

	timer = g_timeout_add_seconds_full(G_PRIORITY_HIGH, WATCHDOG_REFRESH_TIME, watchdog_cb, NULL, NULL);
	if (timer) {
		ret = aw_register(WATCHDOG_TIMEOUT);
		if (ret < 0)
			_E("aw_register failed.");
	}

	/* g_main_loop */
	CRITICAL_LOG("Starting deviced.");
	g_main_loop_run(mainloop);
	g_main_loop_unref(mainloop);

	devices_exit(NULL);
	unload_plugins();

	return 0;
}

int main(int argc, char **argv)
{
	writepid(PIDFILE_PATH);
	return deviced_main(argc, argv);
}
