/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <limits.h>

#include <glib.h>
#include <hw/cpu.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/devices.h"
#include "core/common.h"
#include "core/list.h"
#include "core/device-notifier.h"

#define DEFAULT_PMQOS_TIMER		3000

#define MILLISECONDS(tv)	((tv.tv_sec)*1000 + (tv.tv_nsec)/1000000)

struct pmqos_cpu {
	char* name;
	long timeout;
	guint timeout_id;
};

static dd_list *pmqos_head;
static struct cpu_device *cpu_dev;

static long get_current_monotonic_ms(void)
{
	struct timespec current;
	clock_gettime(CLOCK_MONOTONIC, &current);

	return MILLISECONDS(current);
}

int set_cpu_pmqos(const char *name, int val)
{
	if (!cpu_dev ||
		!cpu_dev->start_boost ||
		!cpu_dev->stop_boost) {
		_E("Cpu device HAL does not exist.");
		return -ENODEV;
	}

	if (!name)
		return -EINVAL;

	if (val) {
		_D("Set pm scenario. Lock %s.", name);
		return cpu_dev->start_boost((void *)name);
	}

	_D("Set pm scenario. Unlock %s.", name);
	return cpu_dev->stop_boost((void *)name);
}

static int pmqos_cpu_cancel(const char *name)
{
	dd_list *elem;
	struct pmqos_cpu *cpu;
	size_t len;

	if (!name)
		return -EINVAL;

	/* Find previous request */
	len = strlen(name) + 1;
	DD_LIST_FOREACH(pmqos_head, elem, cpu) {
		if (!strncmp(cpu->name, name, len))
			break;
	}
	/* no cpu */
	if (!cpu)
		return 0;

	if (cpu->timeout_id) {
		g_source_remove(cpu->timeout_id);
		cpu->timeout_id = 0;
	}

	/* unlock cpu */
	set_cpu_pmqos(cpu->name, false);

	return 0;
}

static gboolean pmqos_cpu_timeout(void *data)
{
	struct pmqos_cpu *cpu = (struct pmqos_cpu *)data;

	cpu->timeout_id = 0;
	set_cpu_pmqos(cpu->name, false);

	return G_SOURCE_REMOVE;
}

static int pmqos_cpu_request(const char *name, int val)
{
	dd_list *elem;
	struct pmqos_cpu *cpu;
	size_t len;
	long next_timeout;

	/* Check valid parameter */
	if (val > DEFAULT_PMQOS_TIMER) {
		_I("The timer value cannot be higher than default time value(%dms).", DEFAULT_PMQOS_TIMER);
		val = DEFAULT_PMQOS_TIMER;
	}

	/* find cpu */
	len = strlen(name) + 1;
	DD_LIST_FOREACH(pmqos_head, elem, cpu) {
		if (!strncmp(cpu->name, name, len))
			break;
	}

	/* add cpu */
	if (!cpu) {
		cpu = calloc(1, sizeof(struct pmqos_cpu));
		if (!cpu)
			return -ENOMEM;

		cpu->name = strndup(name, NAME_MAX);
		cpu->timeout_id = 0;
		cpu->timeout = 0;
		DD_LIST_APPEND(pmqos_head, cpu);
	}

	next_timeout = get_current_monotonic_ms() + val;
	if (!cpu->timeout_id) { // start new one
		cpu->timeout = next_timeout;
		cpu->timeout_id = g_timeout_add(val, pmqos_cpu_timeout, cpu);
		set_cpu_pmqos(cpu->name, true);
	} else if (next_timeout > cpu->timeout) { // just extend timeout
		g_source_remove(cpu->timeout_id);
		cpu->timeout = next_timeout;
		cpu->timeout_id = g_timeout_add(val, pmqos_cpu_timeout, cpu);
	}

	return 0;
}

static int pmqos_lowbat(void *data)
{
	return set_cpu_pmqos("LowBattery", DATA_VALUE_INT(data));
}

static int pmqos_poweroff(void *data)
{
	return set_cpu_pmqos("PowerOff", DATA_VALUE_INT(data));
}

static GVariant * dbus_pmqos_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int val, ret;

	g_variant_get(param, "(i)", &val);

	if (val < 0) {
		ret = -EINVAL;
		goto error;
	}

	_I("Request pmqos name: %s, val %d", name, val);

	if (val)
		ret = pmqos_cpu_request(name, val);
	else
		ret = pmqos_cpu_cancel(name);

error:
	return g_variant_new("(i)", ret);
}

static GVariant * dbus_wifi_pmqos_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char names[NAME_MAX];
	int bps, val, ret;

	g_variant_get(param, "(ii)", &bps, &val);

	if (bps < 0 || val < 0) {
		ret = -EINVAL;
		goto error;
	}

	/* combine bps and member */
	snprintf(names, sizeof(names), "%s%d", name, bps);

	if (val)
		ret = pmqos_cpu_request(names, val);
	else
		ret = pmqos_cpu_cancel(names);

error:
	return g_variant_new("(i)", ret);
}

static GVariant * dbus_getdefaultlocktime(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = DEFAULT_PMQOS_TIMER;

	return g_variant_new("(i)", ret);
}

static GVariant * dbus_psmode_pmqos_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, val;

	g_variant_get(param, "(i)", &val);

	ret = set_cpu_pmqos("UltraPowerSaving", val);

	return g_variant_new("(i)", ret);
}

/* Add pmqos name as alphabetically */
static const dbus_method_s dbus_methods[] = {
	{ "AppLaunch",            "i",    "i", dbus_pmqos_handler },
	{ "AppLaunchHome",        "i",    "i", dbus_pmqos_handler },
	{ "BeautyShot",           "i",    "i", dbus_pmqos_handler },
	{ "Browser",              "i",    "i", dbus_pmqos_handler },
	{ "BrowserDash",          "i",    "i", dbus_pmqos_handler },
	{ "BrowserJavaScript",    "i",    "i", dbus_pmqos_handler },
	{ "BrowserLoading",       "i",    "i", dbus_pmqos_handler },
	{ "BrowserScroll",        "i",    "i", dbus_pmqos_handler },
	{ "CallSound",            "i",    "i", dbus_pmqos_handler },
	{ "CameraBurstShot",      "i",    "i", dbus_pmqos_handler },
	{ "CameraCaptureAtRec",   "i",    "i", dbus_pmqos_handler },
	{ "CameraPreview",        "i",    "i", dbus_pmqos_handler },
	{ "CameraRecording",      "i",    "i", dbus_pmqos_handler },
	{ "CameraSoundAndShot",   "i",    "i", dbus_pmqos_handler },
	{ "ContactSearch",        "i",    "i", dbus_pmqos_handler },
	{ "Emergency",            "i",    "i", dbus_pmqos_handler },
	{ "GalleryScroll",        "i",    "i", dbus_pmqos_handler },
	{ "GalleryRotation",      "i",    "i", dbus_pmqos_handler },
	{ "GetDefaultLockTime",  NULL,    "i", dbus_getdefaultlocktime },
	{ "GpsSerialCno",         "i",    "i", dbus_pmqos_handler },
	{ "GpuBoost",             "i",    "i", dbus_pmqos_handler },
	{ "GpuWakeup",            "i",    "i", dbus_pmqos_handler },
	{ "HomeScreen",           "i",    "i", dbus_pmqos_handler },
	{ "ImageViewer",          "i",    "i", dbus_pmqos_handler },
	{ "IMEInput",             "i",    "i", dbus_pmqos_handler },
	{ "LockScreen",           "i",    "i", dbus_pmqos_handler },
	{ "LowBattery",           "i",    "i", dbus_pmqos_handler },
	{ "MtpSendFile",          "i",    "i", dbus_pmqos_handler },
	{ "MusicPlayLcdOn",       "i",    "i", dbus_pmqos_handler },
	{ "PowerSaving",          "i",    "i", dbus_pmqos_handler },
	{ "ProcessCrashed",       "i",    "i", dbus_pmqos_handler },
	{ "ReservedMode",         "i",    "i", dbus_pmqos_handler },
	{ "ScreenMirroring",      "i",    "i", dbus_pmqos_handler },
	{ "SmemoZoom",            "i",    "i", dbus_pmqos_handler },
	{ "SVoice",               "i",    "i", dbus_pmqos_handler },
	{ "WebappLaunch",         "i",    "i", dbus_pmqos_handler },
	{ "WifiThroughput",      "ii",    "i", dbus_wifi_pmqos_handler },
	{ "PowerOff",             "i",    "i", dbus_pmqos_handler },
	{ "WebAppDrag",           "i",    "i", dbus_pmqos_handler },
	{ "WebAppFlick",          "i",    "i", dbus_pmqos_handler },
	{ "SensorWakeup",         "i",    "i", dbus_pmqos_handler },
	{ "UgLaunch",             "i",    "i", dbus_pmqos_handler },
	{ "MusicScroll",          "i",    "i", dbus_pmqos_handler },
	{ "FileScroll",           "i",    "i", dbus_pmqos_handler },
	{ "VideoScroll",          "i",    "i", dbus_pmqos_handler },
	{ "EmailScroll",          "i",    "i", dbus_pmqos_handler },
	{ "ContactScroll",        "i",    "i", dbus_pmqos_handler },
	{ "TizenStoreScroll",     "i",    "i", dbus_pmqos_handler },
	{ "CallLogScroll",        "i",    "i", dbus_pmqos_handler },
	{ "MyfilesScroll",        "i",    "i", dbus_pmqos_handler },
	{ "UltraPowerSaving",     "i",    "i", dbus_psmode_pmqos_handler },
	{ "WirelessSendFile",     "i",    "i", dbus_pmqos_handler },
	{ "BixbyWakeup",          "i",    "i", dbus_pmqos_handler },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_PMQOS,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	static int done;
	int ret = 0;

	if (data == NULL)
		goto out;

	done = *(int *)data;
	if (!done)
		goto out;
	_I("Booting done.");

	ret = dbus_handle_register_dbus_object_all(NULL);
	if (ret < 0)
		_E("Failed to register dbus method: %d", ret);

	/* register notifier for each event */
	register_notifier(DEVICE_NOTIFIER_CPU_BOOST_LOWBAT, pmqos_lowbat);
	register_notifier(DEVICE_NOTIFIER_CPU_BOOST_POWEROFF, pmqos_poweroff);

out:
	return done;
}

static int pmqos_probe(void *data)
{
	struct hw_info *info;
	int r;

	if (cpu_dev)
		return 0;

	r = hw_get_info(CPU_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (r < 0) {
		_I("Cpu shared library is not supported: %d", r);
		return -ENODEV;
	}

	if (!info->open) {
		_E("Failed to open cpu device: open(NULL)");
		return -EPERM;
	}

	r = info->open(info, NULL, (struct hw_common **)&cpu_dev);
	if (r < 0) {
		_E("Failed to get cpu device structure: %d", r);
		return -EPERM;
	}

	_D("Cpu device structure load success.");
	return 0;
}

static void pmqos_init(void *data)
{
	int ret;

	/* register dbus methods */
	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_PMQOS, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void pmqos_exit(void *data)
{
	dd_list *elem, *elem_next;
	struct pmqos_cpu *cpu;

	DD_LIST_FOREACH_SAFE(pmqos_head, elem, elem_next, cpu) {
		DD_LIST_REMOVE(pmqos_head, cpu);
		free(cpu->name);
		free(cpu);
	}

	/* unregister notifier for each event */
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	unregister_notifier(DEVICE_NOTIFIER_CPU_BOOST_LOWBAT, pmqos_lowbat);
	unregister_notifier(DEVICE_NOTIFIER_CPU_BOOST_POWEROFF, pmqos_poweroff);
}

static const struct device_ops pmqos_device_ops = {
	DECLARE_NAME_LEN("pmqos"),
	.probe    = pmqos_probe,
	.init     = pmqos_init,
	.exit     = pmqos_exit,
};

DEVICE_OPS_REGISTER(&pmqos_device_ops)
