/*
 * deviced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <fcntl.h>
#include <gio/gio.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <stdbool.h>

#include "plugin.h"
#include "common.h"
#include "core/log.h"

#ifndef LIBPATH
#error LIBPATH is not defined.
#endif
#define MODULE_PATH		LIBPATH"/deviced/"

static struct display_plugin disp_plgn;
static struct battery_plugin battery_plgn;
static GList *plgn_list;

inline struct display_plugin *get_display_plugin(void)
{
	return &disp_plgn;
}

inline struct battery_plugin *get_battery_plugin(void)
{
	return &battery_plgn;
}

int load_plugin(const char *id, void **h)
{
	char path[PATH_MAX];
	void *handle;

	if (!id || !h) {
		_E("Wrong parameters");
		return -EINVAL;
	}

	/* Find matched module path */
	snprintf(path, sizeof(path), "%s%s", MODULE_PATH, id);
	if (access(path, R_OK) != 0) {
		_E("There is no %s device", id);
		return -ENODEV;
	}

	/* Load module */
	handle = dlopen(path, RTLD_NOW|RTLD_GLOBAL);
	if (!handle) {
		_E("Failed to open module: %s", dlerror());
		return -ENOENT;
	}

	*h = handle;

	return 0;
}

int unload_plugin(void *h)
{
	if (!h)
		return 0;
	else
		return dlclose(h);
}

void load_plugins()
{
	DIR *dir;
	void *handle = NULL;
	struct dirent *ent;

	if ((dir = opendir(MODULE_PATH)) == NULL) {
		_I("There is no plugin.");
		return ;
	}

	while ((ent = readdir(dir)) != NULL) {
		// skip (/.)(/..)(.so)
		if (strlen(ent->d_name) < 4)
			continue;

		if (strcmp((ent->d_name + strlen(ent->d_name)-3), ".so"))
			continue;

		if (!load_plugin(ent->d_name, &handle)) {
			_I("Plugin(%s) is loaded. handle=%#x", ent->d_name, (unsigned int)((intptr_t)handle));
			plgn_list = g_list_append(plgn_list, handle);
			if (!strcmp(ent->d_name, "display"))
				disp_plgn.handle = handle;
			else if (!strcmp(ent->d_name, "battery"))
				battery_plgn.handle = handle;
		}
	}
	closedir(dir);

	return;
}

void unload_plugins()
{
	GList *plugin = plgn_list;
	void *handle = NULL;

	for ( ; plugin != NULL; plugin = g_list_next(plugin)) {
		handle = (void *)plugin->data;
		_I("Plugin(%#x) is unloaded.", (unsigned int)((intptr_t)handle));
		unload_plugin(handle);
	}

	g_list_free(plgn_list);
}
